/*
*	This was written by Arun Sagar and is placed into the public domain.
*	The author hereby disclaims copyright to this source code.
*
*	Pay it forward. 
*/

#include <CCore>


#include "RBoard.h"
#include "RFBoard.h"




class cRFTable
{
private:
	static void StuffCaptureData(cRFastBoard::CaptureArray, long, long*);

	static cRFastBoard::CaptureFlag LineCaptureFlag(cReversiBoard::LongLine rLine);
	static cRFastBoard::StabilityData LineStability(cReversiBoard::LongLine);
	static void LineCaptureValues(cReversiBoard::LongLine rLine, cReversiBoard::Position position, unsigned *pLeft, unsigned *pRight);

	static ushort RLineBase3Value(cReversiBoard::LongLine rl)
	{
		ushort val = 0;
		ulong multiplier=1;

		while(rl)
		{
			unsigned digit;

			digit = (rl&ABuildOverlayMask(cRFastBoard::kPieceBits,0)) - cReversiBoard::kPieceEdge;
			if(digit <= cReversiBoard::kPiece2)
				val+=digit*multiplier;

			multiplier *= 3;
			rl >>=cRFastBoard::kPieceBits;
		}
		return val;
	}
 
	static void DecomposeToBase3(long number, long nDigits, long* rgDigits)
	{
		while (nDigits--)
		{
			rgDigits[nDigits] = cReversiBoard::kPieceEdge + (number%3);
			number /= 3;
		}
	}

	friend inline ulong CountBits(long i) {ulong count=0; while(i) {if (i&1) count++; i>>=1;} return count;}


public:
	static void Fill(cRFastBoard::CaptureArray*);
	static void Fill(cRFastBoard::CaptureFlag*);
	static void FillXCaptureY(cRFastBoard::XCaptureFlag*, cReversiBoard::Position);
	static void FillStability(cRFastBoard::StabilityData*);
	static void FillBase3(cRFastBoard::Base3TableEntry rgb3[])
	{
		ulong count = 1L<<8;

		Assert(rgb3);

		while(count--)
			rgb3[count] = (cRFastBoard::Base3TableEntry)RLineBase3Value((cReversiBoard::LongLine)count);
	}
	static void FillBitCount(ulong *rgBits)
	{
		long i=256;

		while(i--)
		{
			rgBits[i] = CountBits(i);
		}
	}
};


bool cRFastBoard::
Initialize()
{
	cReversiBoard::Position i;


	for(i=0;i<8;i++)
		cRFTable::FillXCaptureY(cRFastBoard::grgXCaptureY[i], (cRFastBoard::Position)i);

	cRFTable::FillBase3(cReversiBoard::grgBase3Table);
	cRFTable::Fill(cReversiBoard::grgCaptureArray);
	cRFTable::Fill(cReversiBoard::grgCaptureFlags);
	cRFTable::FillStability(cRFastBoard::grgStability);
	cRFTable::FillBitCount(cReversiBoard::grgBitCount);

#ifdef DUMPANALYSIS
	if (cRFastBoard::fDumpMoveAnalysis && cRFastBoard::ofsDump.good())
	{
		if (!cRFastBoard::ofsDump.is_open())
		{
			char szTempFile[512];

			AGetTempDirectory(szTempFile, sizeof(szTempFile));
			strcat(szTempFile, cRFastBoard::kszDefaultDumpFile);
			cRFastBoard::ofsDump.open(szTempFile);
		}

		if (cRFastBoard::ofsDump.good())
		{
			cRFastBoard::ofsDump.seekp(ios::beg);
			cRFastBoard::ofsDump << AComputeKIPS() << " kOps/sec." << endl;
		}
	}
#endif
	return false;
}



cRFastBoard::CaptureFlag cRFTable::
LineCaptureFlag(cReversiBoard::LongLine rLine)
{
	typedef enum
	{
		stStart,
		stBeginNone,
		stBeginPiece1,
		stCountLeft,
		stCountRight
	} State;

	cRFastBoard::CaptureFlag flags=0;
	long i;
	State state = stStart;
	long left,right;

	for (i=0; i<8;i++)
	{
		cReversiBoard::Piece p = cReversiBoard::Piece(rLine & ABuildOverlayMask(cRFastBoard::kPieceBits, 0));

		rLine >>= cRFastBoard::kPieceBits;

		switch(state)
		{
			case stStart:
				if ((cReversiBoard::kPieceNone==p)||(cReversiBoard::kPieceEdge==p))
					state = stBeginNone;
				else if(cReversiBoard::kPiece1 == p)
					state = stBeginPiece1;

			break;

			case stBeginNone:
				if(cReversiBoard::kPiece1 == p)
					state = stBeginPiece1;
				else if (cReversiBoard::kPiece2 == p)
				{
					state = stCountRight;
					right = i-1;
				}
			break;

			case stCountRight:
				if ((cReversiBoard::kPieceNone==p)||(cReversiBoard::kPieceEdge==p))
					state = stBeginNone;
				else if (cReversiBoard::kPiece1 == p)
				{
					state = stBeginPiece1;
					flags |= ABuildBitMask(right);
				}
			break;

			case stBeginPiece1:
				if((cReversiBoard::kPieceNone==p)||(cReversiBoard::kPieceEdge==p))
					state = stBeginNone;
				else if (cReversiBoard::kPiece2 == p)
				{
					state = stCountLeft;
					left = i;
				}
			break;

			case stCountLeft:
				if (cReversiBoard::kPiece1 == p)
					state = stBeginPiece1;
				else if ((cReversiBoard::kPieceNone==p)||(cReversiBoard::kPieceEdge==p))
				{
					state = stBeginNone;
					flags |=  ABuildBitMask(i);
				}
			break;

			default:
				Assert(false);
			break;
		}
	}

	return flags;
}

void cRFTable::
Fill(cRFastBoard::CaptureFlag* rgcf)
{
	long count = 1L<<16;

	Assert(rgcf);

	while(count--)
		rgcf[count] = LineCaptureFlag((cReversiBoard::LongLine)count);
}

#define EXTRACTHIPIECE(rl) ((rl)&ABuildOverlayMask(cRFastBoard::kPieceBits, 7*cRFastBoard::kPieceBits))
#define ISHIPIECE(p,kp) (((ulong)kp << (cRFastBoard::kPieceBits*(8-1))) == (p))
#define ISNOTHIPIECE(p,kp) (((ulong)kp << (cRFastBoard::kPieceBits*(8-1))) != (p))


void cRFTable::
LineCaptureValues(cReversiBoard::LongLine rLine, cReversiBoard::Position position, unsigned *pLeft, unsigned *pRight)
{
	long i;
	
	*pLeft = *pRight = 0;

	rLine <<= (position+1)*cRFastBoard::kPieceBits;
	i = 8;
	while(i--)
	{
		long p = rLine & ABuildOverlayMask(cRFastBoard::kPieceBits, 14);
		
		if (ISNOTHIPIECE(p, cReversiBoard::kPiece2))
		{
			if (ISHIPIECE(p, cReversiBoard::kPiece1))
				*pRight = 7-i;
			break;
		}

		rLine <<= cRFastBoard::kPieceBits;
	}

	rLine >>= (*pRight+2)*cRFastBoard::kPieceBits;
	i = 0;
	do
	{
		long p = rLine & ABuildOverlayMask(cRFastBoard::kPieceBits, 14);
		
		if (ISNOTHIPIECE(p, cReversiBoard::kPiece2))
		{
			if (ISHIPIECE(p, cReversiBoard::kPiece1))
				*pLeft = i;
			break;
		}

		rLine >>=cRFastBoard::kPieceBits;
	}
	while(++i < 8);

}

cRFastBoard::StabilityData cRFTable::
LineStability(cReversiBoard::LongLine rLine)
{
	typedef enum
	{
		stEatEdge,
		stEatNone,
		stEatPiece2,
		stCountPiece1,
		stCountBetweenPiece2,
		stCountBetweenNone,
		stFinished
	} State;

	cRFastBoard::StabilityData stability=0, tempstability=0;
	State state = stEatEdge;
	long i;

	for (i=0; i<8;i++)
	{
		cReversiBoard::Piece p = cReversiBoard::Piece(rLine & ABuildOverlayMask(cRFastBoard::kPieceBits, 0));

		rLine >>= cRFastBoard::kPieceBits;

		switch(state)
		{
			case stEatEdge:
				if (cReversiBoard::kPiece1==p)
				{
					state = stCountPiece1;
					stability++;
				}
				else if(cReversiBoard::kPieceNone == p)
					state = stEatNone;
				else if(cReversiBoard::kPiece2 == p)
					state = stEatPiece2;

			break;

			case stCountPiece1:
			{
				if(cReversiBoard::kPieceEdge == p)
					goto done;
				else if (cReversiBoard::kPieceNone == p)
					state = stEatNone;
				else if (cReversiBoard::kPiece2 == p)
					state = stEatPiece2;
				else
					stability++;

			}
			break;

			case stEatNone:
				if (cReversiBoard::kPieceEdge == p)
					goto done;
				else if (cReversiBoard::kPiece2 == p)
				{
					state = stEatPiece2;
					tempstability = 0;
				}
				else if (cReversiBoard::kPiece1 == p)
				{
					tempstability = 1;
					state = stCountBetweenNone;
				}

			break;

			case stEatPiece2:
				if (cReversiBoard::kPieceEdge == p)
					goto done;
				else if (cReversiBoard::kPieceNone == p)
				{
					state = stEatNone;
					tempstability = 0;
				}
				else if (cReversiBoard::kPiece1 == p)
				{
					tempstability = 1;
					state = stCountBetweenPiece2;
				}
			break;

			case stCountBetweenNone:
				if (cReversiBoard::kPieceEdge == p)
					goto done;
				else if (cReversiBoard::kPieceNone == p)
				{
					state = stEatNone;
					//stability += tempstability;
					tempstability = 0;
				}
				else if (cReversiBoard::kPiece2 == p)
				{
					state = stEatPiece2;
					tempstability = 0;
				}
				else
					tempstability++;
			break;
			
			case stCountBetweenPiece2:
				if (cReversiBoard::kPieceEdge == p)
					goto done;
				else if (cReversiBoard::kPiece2 == p)
				{
					state = stEatPiece2;
					stability += tempstability;
					tempstability = 0;
				}
				else if (cReversiBoard::kPieceNone == p)
				{
					state = stEatNone;
					tempstability = 0;
				}
				else
					tempstability++;
			break;

			default:
				Assert(false);
			break;
		}
	}

done:
	return stability + tempstability;
}

void cRFTable::
FillStability(cRFastBoard::StabilityData* rgsd)
{
	long count = 1L<<16;

	Assert(rgsd);

	while(count--)
	{
		rgsd[count] = LineStability((cReversiBoard::LongLine)count) - LineStability(cRFastBoard::InvertPieces((cReversiBoard::LongLine)count));
#ifdef DEBUGELEPHANT
		{

			if (count%7)
				;//continue;

			static char rgc[]={'.','_','x', 'y'};
			long j;
			for(j=0;j<8;j++)
				ATrace("%c", rgc[(count>>(j*cRFastBoard::kPieceBits))&3]);
			ATraceLn(":%d", rgsd[count]);
		}
#endif
	}
}

void cRFTable::
Fill(cRFastBoard::CaptureArray* rgrga)
{
	long i = 81*81;
	const long nDigits = 8;
	long rgDigits[nDigits];

	Assert(rgrga);

	while(i--)
	{
		DecomposeToBase3(i,sizeof(rgDigits)/sizeof(long), rgDigits);
		
		StuffCaptureData(rgrga[i], nDigits, rgDigits);

#ifdef DEBUGELEPHANT
		if (i%7)
			continue;

		static char rgc[]={'_','_','x', 'y'};
		long j;
		for(j=0;j<nDigits;j++)
			ATrace("%c", rgc[rgDigits[j]]);
		ATrace(":");

		for(j=0;j<nDigits;j++)
			ATrace("%d.%d|", rgrga[i][j].left, rgrga[i][j].right);
		ATraceLn("");
#endif
	}

}

void cRFTable::
StuffCaptureData(cRFastBoard::CaptureArray rgrca, long nDigits, long* rgDigits)
{
	typedef enum
	{
		stStart,
		stBeginNone,
		stBeginPiece1,
		stCountLeft,
		stCountRight
	} State;

	long i =0;
	State state = stStart;
	long left,right;

	while (i < nDigits)
	{
		long p = rgDigits[i];

		rgrca[i].left = 0;
		rgrca[i].right = 0;
		switch(state)
		{
			case stStart:
				if ((cReversiBoard::kPieceNone==p)||(cReversiBoard::kPieceEdge==p))
					state = stBeginNone;
				else if(p==cReversiBoard::kPiece1)
					state = stBeginPiece1;

			break;

			case stBeginNone:
				if (p==cReversiBoard::kPiece2)
					state = stCountRight;
				else if(p==cReversiBoard::kPiece1)
					state = stBeginPiece1;
				right = i-1;
			break;

			case stCountRight:
				if ((cReversiBoard::kPieceNone==p)||(cReversiBoard::kPieceEdge==p))
					state = stBeginNone;
				else if (p==cReversiBoard::kPiece1)
				{
					state = stBeginPiece1;
					rgrca[right].right =  i-right-1;
				}
			break;

			case stBeginPiece1:
				if((cReversiBoard::kPieceNone==p)||(cReversiBoard::kPieceEdge==p))
					state = stBeginNone;
				else if (p == cReversiBoard::kPiece2)
					state = stCountLeft;
				left = i-1;
			break;

			case stCountLeft:
				if (p == cReversiBoard::kPiece1)
					state = stBeginPiece1;
				else if ((cReversiBoard::kPieceNone==p)||(cReversiBoard::kPieceEdge==p))
				{
					state = stBeginNone;
					rgrca[i].left =  i-left-1;
				}
			break;

			default:
				Assert(false);
			break;
		}

		i++;
	}
}

void cRFTable::
FillXCaptureY(cRFastBoard::XCaptureFlag* rgxcf, cReversiBoard::Position column)
{
	long i;

	Assert(column<8);
	Assert(rgxcf);

	for(i=0;i<256;i++)
	{
#define FXCY(c) rgxcf[i][c] = (((i<<(c))>>column) & (cRFastBoard::CaptureFlag)ABuildOverlayMask(1, 7-column))
		FXCY(0);
		FXCY(1);
		FXCY(2);
		FXCY(3);
		FXCY(4);
		FXCY(5);
		FXCY(6);
		FXCY(7);
#undef FXCY
	}
}
